@@ -260,7 +260,8 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 	ret = getaddrinfo(v->val, NULL, &hints, &cni);
 	if (ret) {
 		error(logopt, MODPREFIX
-		      "hostname lookup failed: %s\n", gai_strerror(ret));
+		      "hostname lookup for %s failed: %s\n",
+		      v->val, gai_strerror(ret));
 		goto out;
 	}
 
@@ -270,7 +271,8 @@ static int match_my_name(unsigned int logopt, const char *name, struct substvar
 	ret = getaddrinfo(name, NULL, &hints, &ni);
 	if (ret) {
 		error(logopt, MODPREFIX
-		      "hostname lookup failed: %s\n", gai_strerror(ret));
+		      "hostname lookup for %s failed: %s\n",
+		      name, gai_strerror(ret));
 		freeaddrinfo(cni);
 		goto out;
 	}
@@ -645,6 +647,18 @@ static void update_with_defaults(struct amd_entry *defaults,
 		}
 	}
 
+	if (!entry->sublink) {
+		if (defaults->sublink) {
+			tmp = strdup(defaults->sublink);
+			if (tmp)
+				entry->sublink = tmp;
+		} else {
+			v = macro_findvar(sv, "sublink", 2);
+			if (v)
+				entry->sublink = strdup(v->val);
+		}
+	}
+
 	return;
 }
 
@@ -666,7 +680,8 @@ static char *normalize_hostname(unsigned int logopt, const char *host,
 		ret = getaddrinfo(host, NULL, &hints, &ni);
 		if (ret) {
 			error(logopt, MODPREFIX
-			      "hostname lookup failed: %s", gai_strerror(ret));
+			      "hostname lookup for %s failed: %s",
+			      host, gai_strerror(ret));
 			return NULL;
 		}
 		name = strdup(ni->ai_canonname);
@@ -725,6 +740,17 @@ static struct substvar *expand_entry(struct autofs_point *ap,
 		entry->rhost = host;
 	}
 next:
+	if (entry->pref) {
+		if (expand_selectors(ap, entry->pref, &expand, sv)) {
+			debug(logopt, MODPREFIX
+			      "pref expand(\"%s\") -> %s",
+			      entry->pref, expand);
+			free(entry->pref);
+			entry->pref = expand;
+		}
+		sv = macro_addvar(sv, "pref", 4, entry->pref);
+	}
+
 	if (entry->sublink) {
 		if (expand_selectors(ap, entry->sublink, &expand, sv)) {
 			debug(logopt, MODPREFIX
@@ -967,8 +993,8 @@ static int do_auto_mount(struct autofs_point *ap, const char *name,
 static int do_link_mount(struct autofs_point *ap, const char *name,
 			 struct amd_entry *entry, unsigned int flags)
 {
-	char target[PATH_MAX + 1];
 	const char *opts = (entry->opts && *entry->opts) ? entry->opts : NULL;
+	char *target;
 	int ret;
 
 	if (entry->sublink) {
@@ -977,14 +1003,14 @@ static int do_link_mount(struct autofs_point *ap, const char *name,
 			     "error: sublink option length is too long");
 			return 0;
 		}
-		strcpy(target, entry->sublink);
+		target = entry->sublink;
 	} else {
 		if (strlen(entry->fs) > PATH_MAX) {
 			error(ap->logopt, MODPREFIX
 			     "error: fs option length is too long");
 			return 0;
 		}
-		strcpy(target, entry->fs);
+		target = entry->fs;
 	}
 
 	if (!(flags & CONF_AUTOFS_USE_LOFS))
@@ -1057,7 +1083,14 @@ static int do_generic_mount(struct autofs_point *ap, const char *name,
 			umount = 1;
 		}
 		/* We have an external mount */
-		ext_mount_add(&entry->ext_mount, entry->fs, umount);
+		if (!ext_mount_add(&entry->ext_mount, entry->fs, umount)) {
+			umount_ent(ap, entry->fs);
+			error(ap->logopt, MODPREFIX
+			      "error: could not add external mount %s",
+			      entry->fs);
+			ret = 1;
+			goto out;
+		}
 		ret = do_link_mount(ap, name, entry, flags);
 	}
 out:
@@ -1076,7 +1109,7 @@ static int do_nfs_mount(struct autofs_point *ap, const char *name,
 	if (strlen(entry->rhost) + strlen(entry->rfs) + 1 > PATH_MAX) {
 		error(ap->logopt, MODPREFIX
 		     "error: rhost + rfs options length is too long");
-		return 0;
+		return 1;
 	}
 
 	strcpy(target, entry->rhost);
@@ -1101,7 +1134,13 @@ static int do_nfs_mount(struct autofs_point *ap, const char *name,
 			umount = 1;
 		}
 		/* We might be using an external mount */
-		ext_mount_add(&entry->ext_mount, entry->fs, umount);
+		if (!ext_mount_add(&entry->ext_mount, entry->fs, umount)) {
+			umount_ent(ap, entry->fs);
+			error(ap->logopt, MODPREFIX
+			      "error: could not add external mount %s", entry->fs);
+			ret = 1;
+			goto out;
+		}
 		ret = do_link_mount(ap, name, entry, flags);
 	}
 out:
@@ -1286,9 +1325,12 @@ static int do_program_mount(struct autofs_point *ap,
 	 */
 	if (ext_mount_inuse(entry->fs)) {
 		rv = 0;
+		/* An external mount with path entry->fs exists
+		 * so ext_mount_add() won't fail.
+		 */
 		ext_mount_add(&entry->ext_mount, entry->fs, 1);
 	} else {
-		rv = mkdir_path(entry->fs, 0555);
+		rv = mkdir_path(entry->fs, mp_mode);
 		if (rv && errno != EEXIST) {
 			char buf[MAX_ERR_BUF];
 			char *estr;
@@ -1302,17 +1344,19 @@ static int do_program_mount(struct autofs_point *ap,
 
 		rv = spawnv(ap->logopt, prog, (const char * const *) argv);
 		if (WIFEXITED(rv) && !WEXITSTATUS(rv)) {
-			rv = 0;
-			ext_mount_add(&entry->ext_mount, entry->fs, 1);
-			debug(ap->logopt, MODPREFIX
-			      "%s: mounted %s", entry->type, entry->fs);
-		} else {
-			if (!ext_mount_inuse(entry->fs))
-				rmdir_path(ap, entry->fs, ap->dev);
-			error(ap->logopt, MODPREFIX
-			     "%s: failed to mount using: %s",
-			     entry->type, entry->mount);
+			if (ext_mount_add(&entry->ext_mount, entry->fs, 1)) {
+				rv = 0;
+				debug(ap->logopt, MODPREFIX
+				     "%s: mounted %s", entry->type, entry->fs);
+				goto do_free;
+			}
+			umount_ent(ap, entry->fs);
 		}
+
+		if (!ext_mount_inuse(entry->fs))
+			rmdir_path(ap, entry->fs, ap->dev);
+		error(ap->logopt, MODPREFIX
+		   "%s: failed to mount using %s", entry->type, entry->mount);
 	}
 do_free:
 	free_argv(argc, (const char **) argv);
@@ -1827,15 +1871,25 @@ struct amd_entry *make_default_entry(struct autofs_point *ap,
 	if (amd_parse_list(ap, defaults, &dflts, &sv))
 		return NULL;
 	defaults_entry = list_entry(dflts.next, struct amd_entry, list);
-	list_del_init(&defaults_entry->list);
 	/*
 	 * If map type isn't given try to inherit from
 	 * parent. A NULL map type is valid and means
 	 * use configured nss sources.
 	 */
 	map_type = conf_amd_get_map_type(ap->path);
-	if (map_type)
-		defaults_entry->map_type = strdup(map_type);
+	if (map_type) {
+		defaults_entry->map_type = map_type;
+#ifndef HAVE_HESIOD
+		if (!strcmp(map_type, "hesiod")) {
+			warn(ap->logopt, MODPREFIX
+			     "hesiod support not built in, "
+			     "defaults map entry not set");
+			defaults_entry = NULL;
+		}
+#endif
+	}
+	if (defaults_entry)
+		list_del_init(&defaults_entry->list);
 	/* The list should now be empty .... */
 	free_amd_entry_list(&dflts);
 	return defaults_entry;
@@ -1954,15 +2008,27 @@ static struct amd_entry *get_defaults_entry(struct autofs_point *ap,
 			goto out;
 		}
 		entry = select_default_entry(ap, &dflts, sv);
-		if (!entry->map_type) {
+		if (entry && !entry->map_type) {
 			/*
 			 * If map type isn't given try to inherit from
 			 * parent. A NULL map type is valid and means
 			 * use configured nss sources.
 			 */
 			char *map_type = conf_amd_get_map_type(ap->path);
-			if (map_type)
-				entry->map_type = strdup(map_type);
+			if (map_type) {
+				entry->map_type = map_type;
+#ifndef HAVE_HESIOD
+				if (!strcmp(map_type, "hesiod")) {
+					warn(ap->logopt, MODPREFIX
+					     "hesiod support not built in, "
+					     "attempting to use internal "
+					     "default");
+					free_amd_entry(entry);
+					free(expand);
+					goto out;
+				}
+#endif
+			}
 		}
 		free(expand);
 	}
@@ -2092,6 +2158,7 @@ int parse_mount(struct autofs_point *ap, const char *name,
 			list_del_init(&this->list);
 			cur_defaults = this;
 			update_with_defaults(defaults_entry, cur_defaults, sv);
+			debug(ap->logopt, "merged /defaults entry with defaults");
 			continue;
 		} else if (this->flags & AMD_DEFAULTS_RESET) {
 			struct amd_entry *nd, *new;
@@ -2130,8 +2197,9 @@ int parse_mount(struct autofs_point *ap, const char *name,
 
 		at_least_one = 1;
 
-		debug(ap->logopt, "expand mount entry");
+		debug(ap->logopt, "update mount entry with defaults");
 		update_with_defaults(cur_defaults, this, sv);
+		debug(ap->logopt, "expand mount entry");
 		sv = expand_entry(ap, this, flags, sv);
 		sv = merge_entry_options(ap, this, sv);
 		normalize_sublink(ap->logopt, this, sv);
